// batchboxer.C  9/11/01 Steve Ludtke

//N batchboxer
//F This program is used for offline boxing of micrographs (MRC format ONLY!)
//T LM
//1
//P	[dbbox=<filespec>]	Input box database
//P [imgbox=<filespec>]	Read box database from particle image set headers
//P	[auto[=<refs>,<thr1>,<thr2>,<thr3>]]	Autobox with defined thresholds
//P	[xform=<dx>,<dy>,<da>]	Rotate and translate the box database before output (see alignhuge)
//P [newsize=<size>]	New box size for all boxes (center preserved)
//P [input=<filespec>]	Image to be boxed	(MRC format only!!!)
//P	[output=<filespec>]	Output file for boxed images
//P [ctfout=<filespec>]	File to put CTF curves into (if input is focal pair merged)
//P	[dbout=<filespec>]	Output box database file
//U batchboxer dbbox=jj0441.box input=jj0441.mrc output=0441.hed
//D This program does offline boxing of particles using a box database or an automatic
//E boxing routine. Please note that it will work only on MRC images.

#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"

#define CLIP	768

#define DB(x) printf("%s\n",x);

// class for dealing with box databases, the x,y coordinates are 
// stored internally as the center of the box, but returned and
// passed in as the lower left corner of the image
class BOXLIST {
	int *x,*y;
	int n,na;
	int sz;
	
public:
	BOXLIST();
	~BOXLIST();
	void empty();
	void append(int x,int y);
	int atx(int i);
	int aty(int i);
	void setsize(int s);
	int size();
	int count();
	void xform(int dx,int dy,float da,int xsize,int ysize);
	void write(const char *file);
	void read(const char *file);
	void readimg(const char *file);
};

BOXLIST::BOXLIST() { 
	x=(int *)malloc(8*sizeof(int)); 
	y=(int *)malloc(8*sizeof(int));
	na=8;
	n=0;
}

BOXLIST::~BOXLIST() {
	free(x);
	free(y);
}

int BOXLIST::size() { return sz; }

void BOXLIST::setsize(int s) { sz=s; }

void BOXLIST::empty() { n=0; }

void BOXLIST::append(int X, int Y) { 
	x[n]=X+sz/2; 
	y[n]=Y+sz/2;
	n++; 
	if (n>=na) { 
		na*=2;
		x=(int *)realloc(x,na*sizeof(int));
		y=(int *)realloc(y,na*sizeof(int));
	}
}

int BOXLIST::atx(int i) { if (i<n&&i>=0) return x[i]-sz/2; return 0; }
int BOXLIST::aty(int i) { if (i<n&&i>=0) return y[i]-sz/2; return 0; }

int BOXLIST::count() { return n; }

void BOXLIST::write(const char *file) {
FILE *out = fopen(file,"w");
if (!out) return;
for (int i=0; i<n ; i++) fprintf(out,"%d\t%d\t%d\t%d\n",atx(i),aty(i),sz,sz);
fclose(out);
}

void BOXLIST::read(const char *file) {
FILE *in = fopen(file,"r");
if (!in) return;

char s[128];
int xx,yy;
while (fgets(s,127,in)) {
	if (sscanf(s," %d %d %d",&xx,&yy,&sz)!=3) continue;
	append(xx,yy);
}
fclose(in);

}

void BOXLIST::readimg(const char *file) {
int i,j=0,x0,y0,x1,y1;
char t[80];
EMData r;
int n=fileCount(file);

for (i=0; i<n; i++) {
	r.readImage(file,i,1);
	if (sscanf(r.Name()," %d,%d",&x0,&y0)!=2) {
		if (sscanf(r.Name(),"!$ %d,%s %d,%d",&j,t,&x0,&y0)!=4)  continue;
	}
	x1=r.xSize();
	y1=r.ySize();

	append(x0,y0);
}

sz=x1; 
}

void BOXLIST::xform(int dx,int dy,float da,int xs,int ys) {
float mx0=cos(da);
float mx1=sin(da);

for (int i=0; i<n; i++) {
	float tx= (x[i]-xs/2)*mx0+(y[i]-ys/2)*mx1-dx+(xs/2);
	float ty=-(x[i]-xs/2)*mx1+(y[i]-ys/2)*mx0-dy+(ys/2);
	x[i]=tx;
	y[i]=ty;
}

}

BOXLIST boxes;
unsigned char *bestmap=NULL;

void autoBox(char *micro,char *ref,float *thresh);
void boxesFrom(EMData *map,EMData *data,List &abptcl,int x0,int y0,int sx,int sy,float *thresh,int boxsize);
EMData *ptclCorr(char *micro, List &refs,int xo,int yo);

main(int argc,char *argv[]) {
char *input=NULL,*boxed=NULL,*dbout=NULL,*dbin=NULL,*imgin=NULL,*ctfout=NULL;
float autothr[4] = { 0,0,0,1.0 };
float xfdx=0,xfdy=0,xfda=0;
char refs[512];
int i,newsize=0;

appinit(argc,argv);

if (argc<2) appusage(argv[0]);

for (i=1; i<argc; i++) {	
	if (strncmp(argv[i],"input=",6)==0) input=argv[i]+6;
	else if (strncmp(argv[i],"output=",7)==0) boxed=argv[i]+7; 
	else if (strncmp(argv[i],"ctfout=",7)==0) ctfout=argv[i]+7; 
	else if (strncmp(argv[i],"newsize=",8)==0) newsize=atoi(argv[i]+8); 
	else if (strncmp(argv[i],"dbout=",6)==0) dbout=argv[i]+6;
	else if (strncmp(argv[i],"dbbox=",6)==0) dbin=argv[i]+6;
	else if (strncmp(argv[i],"imgbox=",7)==0) imgin=argv[i]+7;
	else if (strncmp(argv[i],"auto=",5)==0) sscanf(argv[i]+5,"%[^,],%f,%f,%f",refs,autothr,autothr+1,autothr+2);
	else if (strncmp(argv[i],"xform=",6)==0) sscanf(argv[i]+6,"%f,%f,%f",&xfdx,&xfdy,&xfda);
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

int Ref=LOGbegin(argc,argv);

boxes.empty();
if (dbin) { printf("Read boxes from %s\n",dbin); boxes.read(dbin); }
if (imgin) { printf("Extract location info from %s\n",imgin); boxes.read(imgin); }
if (autothr[0]||autothr[1]||autothr[2]) {
	autoBox(input,refs,autothr);
}

if (boxes.count()==0) error(ERR_EXIT,"No boxes to process");
printf("%d Boxes to process\n",boxes.count());

if (newsize>0) boxes.setsize(newsize);

if (xfdx||xfdy||xfda) {
	if (!input) error(ERR_EXIT,"Cannot perform transformation without an input image.");
	EMData tmp;
	tmp.readImage(input,0,1);
	boxes.xform(xfdx,xfdy,xfda,tmp.xSize(),tmp.ySize());
}


if (dbout) boxes.write(dbout);
if (boxed) {
	if (!input) error(ERR_EXIT,"Cannot output boxed particles without an input image");
	printf("Boxing %s into %s\n",input,boxed);
	EMData tmp;
	tmp.readMRC(input,1);
	char s[120];
	int ni;
	strcpy(s,tmp.Name());
	for (i=0; i<boxes.count(); i++) {
		tmp.readMRCArea(input,boxes.atx(i),boxes.aty(i),boxes.size(),boxes.size());
		char t[120];
		if (s[0]=='!' && s[1]=='$') {
			if (ctfout) {
				char nm[120];
				float apix;
				int u;
				sscanf(s,"!$ %d,%f,%s",&u,&apix,nm);
				EMData tmp2;
				if (i==0) {
					tmp2.readImage(nm,u);
					tmp2.writeIMAGIC(ctfout,-1);
					ni=fileCount(ctfout)-1;
				}
				sprintf(t,"!$ %d,%1.5g,%s %d,%d",ni,apix,ctfout,boxes.atx(i),boxes.aty(i));
			}
			else sprintf(t,"%s %d,%d",s,boxes.atx(i),boxes.aty(i));
		}
		else sprintf(t,"%d,%d",boxes.atx(i),boxes.aty(i));
		tmp.setName(t);
		tmp.writeIMAGIC(boxed,-1);
	}
}


LOG(Ref,NULL,LOG_END,NULL);
}


void autoBox(char *micro,char *ref,float *thresh) {
EMData *d1,*d2,*data;
int px,py,i;
List *refs;
List abptcl;

data=new EMData;
data->readMRC(micro,1);

refs=readImages(ref,-1,-1);
d1=(EMData *)refs->objectAt(0);
boxes.setsize(d1->ySize());

if (refs->count()<1) return;
if (refs->count()>250) error(ERR_EXIT,"FATAL ERROR: Max # references 250.");

abptcl.empty();
for (i=0; i<refs->count(); i++) {
	d1=(EMData *)refs->objectAt(i);
	d1->edgeNormalize();
//	d1->applyMask(-1,4);
	
	d2=d1->copy(0,0);
	d2->medianShrink(2);
	abptcl.addObject(d2);
	delete d1;
}
delete refs;
int boxsize=((EMData *)abptcl.objectAt(0))->xSize();

//ab->setup(px,py,d1->Min(),d1->Max(),d1,this);

int j,nx,ny;

nx=(data->xSize()-boxsize*2)/(CLIP*2-boxsize)+1;
ny=(data->ySize()-boxsize*2)/(CLIP*2-boxsize)+1;

for (i=0; i<ny; i++) {
	for (j=0; j<nx; j++) {
		printf("%d,%d (%d)\n",i,j,boxes.count());
		d1=ptclCorr(micro,abptcl,j*(CLIP*2-boxsize),i*(CLIP*2-boxsize));
		boxesFrom(d1,data,abptcl,j*(CLIP*2-boxsize),i*(CLIP*2-boxsize),0,0,thresh,boxsize);
		delete d1;
	}
}

}

void boxesFrom(EMData *map,EMData *data,List &abptcl,int x0,int y0,int sx,int sy,float *thresh,int boxsize) {
int i,j,k,l,c,n=0,mn;
float *d,*d2,da1,da2,bd;
EMData *dt,*dtc;

if (!map) return;

if (sx<boxsize/4) sx=boxsize/4;
if (sy<boxsize/4) sy=boxsize/4;

mn=4*CLIP*CLIP*4/(boxsize*boxsize);

//DB("a")
dtc=new EMData;
dtc->readMRCArea(data->Path(),x0-128,y0-128,CLIP*2+256,CLIP*2+256);
d=map->getData();
for (j=sy; j<CLIP-boxsize/4; j++) {
	for (i=sx; i<CLIP-boxsize/4; i++) {
		if (d[i+j*CLIP]>thresh[0]) {
			if (n>mn ||i*2+x0>data->xSize()-boxsize || j*2+y0>data->ySize()-boxsize) continue;
			dt=dtc->clip(i*2,j*2,boxsize,boxsize);
			printf("%f ",d[i+j*CLIP]);

			// Make sure the particle really looks like the reference
			if (thresh[1]<1.0 || thresh[2]>0) {
				EMData *cp=dt->copy(0,0);
				EMData *rf=((EMData *)abptcl.objectAt(bestmap[i+j*CLIP]))->copy(0,0);
//				cp->medianShrink(2);
				float mult,add;
				rf->normalizeTo(cp,0,0,&mult,&add);
				cp->subtract(rf);
				
				// now calculate the Kurtosis of the resulting image (0 for gaussians)
				float mean=cp->Mean(),sig=cp->Sigma(),kurt=0;
				float *d=cp->getDataRO();
				for (l=0; l<boxsize*boxsize/4; l++) kurt+=pow((d[l]-mean)/sig,4.0);
				kurt=kurt/(boxsize*boxsize/4.0)-3.0;
				cp->doneData();

printf("%f %f %f\n",cp->Sigma(),mult,fabs(kurt));
				if (fabs(kurt)>thresh[3] || mult<thresh[2] || cp->Sigma()>thresh[1] || mult==1) { delete cp; delete rf; delete dt; continue; }
				
				delete cp;
				delete rf;
			}

			// ok, we've got a good particle			
			boxes.append(i*2+x0-boxsize/2,j*2+y0-boxsize/2);
			delete dt;
			n++;
		}
	}
}
//DB("b")
delete dtc;
map->doneData();
}

EMData *ptclCorr(char *micro, List &refs,int xo,int yo) { 
EMData *dc,*dd,*ccf,*ref,*dt;
int i,j,ii,jj,k,x1,y1;
float *dp,*ddp,t,m,s;

int boxsize = ((EMData *)refs.objectAt(0))->ySize();

dd=new EMData;
// a little padding to get rid of edge effects
dd->readMRCArea(micro,xo-128,yo-128,CLIP*2+256,CLIP*2+256);
dd->medianShrink(2);

// Now we fill in those padded zero regions with decent values
dp=dd->getDataRO();

dc=dd->clip(64,64,CLIP,6);
m=dc->Mean();
s=dc->Sigma();
delete dc;
for (j=0; j<64; j++) {
	for (i=0; i<CLIP+128; i++) {
		dp[i+j*(CLIP+128)]=grand(m,s);
	}
}

dc=dd->clip(64,64+CLIP-6,CLIP,6);
m=dc->Mean();
s=dc->Sigma();
delete dc;
for (j=CLIP+64; j<CLIP+128; j++) {
	for (i=0; i<CLIP+128; i++) {
		dp[i+j*(CLIP+128)]=grand(m,s);
	}
}

dc=dd->clip(64,64,6,CLIP);
m=dc->Mean();
s=dc->Sigma();
delete dc;
for (j=64; j<64+CLIP; j++) {
	for (i=0; i<64; i++) {
		dp[i+j*(CLIP+128)]=grand(m,s);
	}
}

dc=dd->clip(64+CLIP-6,64,6,CLIP);
m=dc->Mean();
s=dc->Sigma();
delete dc;
for (j=64; j<64+CLIP; j++) {
	for (i=CLIP+64; i<CLIP+128; i++) {
		dp[i+j*(CLIP+128)]=grand(m,s);
	}
}

dd->doneData();
dd->update();

//dd->writeIMAGIC("xxx.hed",-1);

//DB("aa")

dd->filter(3,10000,1);		// high pass filter to clean up gradients
dc=dd->clip(64,64,CLIP,CLIP);
delete dd;

dd=dc->copy(0,0);
dd->zero();

//DB("ab")

if (!bestmap) bestmap=(unsigned char *)malloc(CLIP*CLIP);
else bestmap=(unsigned char *)realloc(bestmap,CLIP*CLIP);
for (i=0; i<refs.count(); i++) {
//	ptcl.at(i)->writeIMAGIC("ref.hed",-1);
	ref=((EMData *)refs.objectAt(i))->clip(-(CLIP-boxsize/2)/2,-(CLIP-boxsize/2)/2,CLIP,CLIP);
	ref->normalize();
	ref->toCorner();
	ccf=dc->calcCCF(ref);

	dp=ccf->getDataRO();
	ddp=dd->getData();
	for (j=0; j<CLIP*CLIP; j++) {
		if (dp[j]>ddp[j]) { ddp[j]=dp[j]; bestmap[j]=i; }
	}
	ccf->doneData();
	dd->doneData();
	
	delete ccf;
	delete ref;
}

//DB("ac")

// This does a low pass filter so are peaks are guarenteed to be separated from each other
dd->filter(0,CLIP/6.0,5);
t=dd->Max();
dp=dd->getData();

// first, we locate peaks
for(i=boxsize/4; i<CLIP-boxsize/4-1; i++) {
	for (j=boxsize/4; j<CLIP-boxsize/4-1; j++) {
		k=i+j*CLIP;
		if (dp[k]>dp[k-1] && dp[k]>dp[k+1] && dp[k]>dp[k+CLIP] && dp[k]>dp[k-CLIP] && dp[k]>0 &&
			dp[k]>dp[k-1-CLIP] && dp[k]>dp[k+1-CLIP] && dp[k]>dp[k-1+CLIP] && dp[k]>dp[k+1+CLIP]) {
				dp[k]+=t;
		}
	}
}

// Now we eliminate boxes which are too close to each other
for(i=boxsize/4; i<CLIP-boxsize/4-1; i++) {
	for (j=boxsize/4; j<CLIP-boxsize/4-1; j++) {
		k=i+j*CLIP;
		if (dp[k]>t) {
			x1=i+boxsize/4-1;
			y1=j+boxsize/4-1;
			for (ii=i-boxsize/4; ii<x1; ii++) {
				for (jj=j-boxsize/4; jj<y1; jj++) {
					if (dp[ii+jj*CLIP]>dp[k]) { dp[k]=0; break; }
				}
			}

		}
	}
}

// now we eliminate everything but the peaks
for (i=1; i<CLIP-1; i++) {
	for (j=1; j<CLIP-1; j++) {
		k=i+j*CLIP;
		if (dp[k]<=t) dp[k]=0; 
		else {
			dp[k]-=t;
			dp[k]*=100.0;
		}
	}
}

//DB("ad")

dd->doneData();

delete dc;
//delete dd;
return dd;
}

