// segment3d.C  4/3/00  Steve Ludtke
//N segment3d
//F Split a 3d map into several pieces.
//T LT
//1
//P <input file>	Source file
//P <output file>	Destination file
//P <n seg>	Number of segments to split map into
//P [ampw]	Apply a straight line integral amplitude weight to the distance parameter
//P [thr=<val>]	Only segment voxels above the given threshold
//P [split]	Separate out the individual segments 
//P [maxit=<n>]	Specify maximum iterations before giving up on convergence
//P [itxp=<n>] After the segmentation, iteratively extend the segments into regions below threshold, n is the number of voxels of expansion
//P [gaussmod=<mrc out>]	Construct a 3d model from Gaussians centered at the group centers.
//P [randomize]	This will randomize the gaussians by rotating them randomly with a fixed radius
//U segment3d threed.1a.mrc seg.mrc 12 thr=1
//D This program will take a 3D map and attempt to intelligently split that map into 'segments'.
//E That is, every voxel in the map will be assigned to one of <n seg> segments. The assignment
//E is performed in such a way that high density regions will be grouped together. This is
//E accomplished using a k-means clustering algorithm.

#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include <time.h>

struct Segment { float x,y,z,s,v,sigx,sigy,sigz; };

main(int argc, char *argv[]) {
int i,j,k,x,y,z,rize=0;
int ampw=0,nseg=0,split=0,mit=999999,it=0,itxp=0;
float thr=-1.0e20;
float h,bh,ra;
float *ds,*dd;
char *gaussmod = NULL,*vector = NULL;
EMData *data,*segmap;
Segment *seg,*oseg,*tseg;

appinit(argc,argv);

if (argc<4) appusage(argv[0]);

nseg=atoi(argv[3]);
if (nseg<2) { printf("Must have at least 2 segments!\n"); exit(1); }
seg=(Segment *)calloc(nseg,sizeof(Segment));
oseg=(Segment *)calloc(nseg,sizeof(Segment));

for (i=4; i<argc; i++) {	
	if (strncmp(argv[i],"ampw",4)==0) ampw=1;
	else if (strncmp(argv[i],"thr=",4)==0) thr=atof(&argv[i][4]);
	else if (strncmp(argv[i],"split",5)==0) split=1;
	else if (strncmp(argv[i],"itxp=",5)==0) itxp=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"maxit=",5)==0) mit=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"vector=",7)==0) vector=&argv[i][7];
	else if (strncmp(argv[i],"gaussmod=",9)==0) gaussmod=&argv[i][9];
	else if (strncmp(argv[i],"randomize",9)==0) rize=1;
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

data=new EMData;
data->readImage(argv[1],-1);

segmap=new EMData;
segmap=data->copyHead();
segmap->zero();

dd=data->getDataRO();
ds=segmap->getData();

int nx=data->xSize();
int ny=data->ySize();
int nz=data->zSize();

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

// Initially we start with random segment positions
srandom(time(0));
for (i=0; i<nseg; i++) {
	seg[i].x=random()%nx;
	seg[i].y=random()%ny;
	seg[i].z=random()%nz;
}

while (it<mit) {
	it++;
	// Now we assign each voxel to a segment based on various distance criteria
	for (z=0,k=0; z<nz; z++) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx; x++,k++) {
				if (dd[k]<=thr) { ds[k]=-1.0; continue; }
				
				bh=nx*ny*nz;	//not important, just big
				// find which segment this voxel goes in
				for (i=0; i<nseg; i++) {
					if (ampw) {
						h=sqrt(hypot3(x-seg[i].x,y-seg[i].y,z-seg[i].z));
						float sum=0;
						for (float rr=0; rr<=1.0; rr+=1.0/h) 
							sum+=data->valueAt(x*rr+seg[i].x*(1.0-rr),y*rr+seg[i].y*(1.0-rr),z*rr+seg[i].z*(1.0-rr)); 
						h=h*h/sum;
						if (h<bh) { bh=h; j=i; }
						
					}
					else {
						h=hypot3(x-seg[i].x,y-seg[i].y,z-seg[i].z);
						if (h<bh) { bh=h; j=i; }
					}
				}
				ds[k]=j;
			}
		}
	}			

	// now we redetermine the center locations
	tseg=seg;
	seg=oseg;
	oseg=tseg;
	memset(seg,0,nseg*sizeof(Segment));
	for (z=0,k=0; z<nz; z++) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx; x++,k++) {
				if (dd[k]>thr && ds[k]>=0) {
					seg[(int)ds[k]].x+=x*dd[k];
					seg[(int)ds[k]].y+=y*dd[k];
					seg[(int)ds[k]].z+=z*dd[k];
					seg[(int)ds[k]].s+=dd[k];	// normalization
					seg[(int)ds[k]].v+=1.0;		// count of voxels in segment
				}
			}
		}
	}			

	for (k=0; k<nseg; k++) {
		if (seg[k].s) {
			seg[k].x/=seg[k].s;
			seg[k].y/=seg[k].s;
			seg[k].z/=seg[k].s;
		}
		printf("%d. %f\t%f\t%f\t%d\n",k,seg[k].x,seg[k].y,seg[k].z,(int)seg[k].v);
		
		if (seg[k].v==0) {
			seg[k].x=random()%nx;
			seg[k].y=random()%ny;
			seg[k].z=random()%nz;
		}
	}

	ra=0;
	for (k=0; k<nseg; k++) {
		ra+=hypot3(seg[k].x-oseg[k].x,seg[k].y-oseg[k].y,seg[k].z-oseg[k].z);
	}
	ra/=(float)nseg;
	printf("%d. --------------   %f\n",it,ra);
	if (ra<.001) break;
}
data->doneData();

if (itxp) {
	float *smc=(float *)malloc(nx*ny*nz*sizeof(float));

	for (i=0; i<itxp; i++) {
		memcpy(smc,ds,nx*ny*nz*sizeof(float));
		for (z=1; z<nz-1; z++) {
			for (y=1; y<ny-1; y++) {
				for (x=1; x<nx-1; x++) {
					k=x+y*nx+z*nx*ny;
					if (ds[k]!=-1.0) continue;
					// if a neighbor has a segment number, use it, if ambiguous, use the largest one (arbitrary)
					ds[k]=smc[k-1];
					if (smc[k+1]>ds[k]) ds[k]=smc[k+1];
					if (smc[k-nx]>ds[k]) ds[k]=smc[k-nx];
					if (smc[k+nx]>ds[k]) ds[k]=smc[k+nx];
					if (smc[k-nx*ny]>ds[k]) ds[k]=smc[k-nx*ny];	
					if (smc[k+nx*ny]>ds[k]) ds[k]=smc[k+nx*ny];
				}
			}
		}
	}	
	free(smc);
}


// This will write the individual segments into individual 3D data files
if (split) {
	int x0,x1,y0,y1,z0,z1;
	char fsp[100];

	for (i=0; i<nseg; i++) {
		EMData *cp = data->copy();
		dd=cp->getData();
		
		// find the limits of this segment and mask it out
		x0=y0=z0=nx+ny+nz;
		x1=y1=z1=0;
		for (z=0,k=0; z<nz; z++) {
			for (y=0; y<ny; y++) {
				for (x=0; x<nx; x++,k++) {
					if (ds[k]==i) {
						if (x<x0) x0=x;
						if (x>x1) x1=x;
						if (y<y0) y0=y;
						if (y>y1) y1=y;
						if (z<z0) z0=z;
						if (z>z1) z1=z;
					}
					else dd[k]=0;
				}	
			}
		}
		cp->doneData();				
		EMData *clip=cp->clip(x0,y0,z0,x1-x0+1,y1-y0+1,z1-z0+1);
		sprintf(fsp,"seg.%d.mrc",i);
		clip->writeMRC(fsp);
		delete clip;
		delete cp;

		printf("Wrote %s   %d x %d x %d\n",fsp,x1-x0+1,y1-y0+1,z1-z0+1);
	}

}

// now write the final segment map
segmap->doneData();

segmap->writeMRC(argv[2]);
LOG(Ref,argv[2],LOG_OUTFILE,NULL);

// Write a text file with the vectors if requested
if (vector) {
	FILE *out = fopen(vector,"w");
	if (out) {
		for (i=0; i<nseg; i++) {
			fprintf(out,"%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\n",seg[i].x,seg[i].y,seg[i].z,seg[i].sigx,seg[i].sigy,seg[i].sigz);
		}
		fclose(out);
	}
}

// write out a map constructed from gaussians at the center of each segment
// note that this destroys data
if (gaussmod) {
	//calculate std dev of each region
	dd=data->getDataRO();
	ds=segmap->getData();
	memset(seg,0,nseg*sizeof(Segment));
	for (z=0,k=0; z<nz; z++) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx; x++,k++) {
				if (dd[k]>thr && ds[k]>=0) {
					seg[(int)ds[k]].x+=x*dd[k];
					seg[(int)ds[k]].y+=y*dd[k];
					seg[(int)ds[k]].z+=z*dd[k];
					seg[(int)ds[k]].s+=dd[k];	// normalization
					seg[(int)ds[k]].v+=1.0;		// count of voxels in segment
					seg[(int)ds[k]].sigx+=x*x*dd[k];
					seg[(int)ds[k]].sigy+=y*y*dd[k];
					seg[(int)ds[k]].sigz+=z*z*dd[k];
				}
			}
		}
	}
	
	for (i=0; i<nseg; i++) {
		seg[i].x/=seg[i].s;
		seg[i].y/=seg[i].s;
		seg[i].z/=seg[i].s;
		seg[i].sigx=seg[i].sigx/seg[i].s-seg[i].x*seg[i].x;
		seg[i].sigy=seg[i].sigy/seg[i].s-seg[i].y*seg[i].y;
		seg[i].sigz=seg[i].sigz/seg[i].s-seg[i].z*seg[i].z;
		seg[i].s=(seg[i].sigx+seg[i].sigy+seg[i].sigz)/3.0;
	}
	data->doneData();
	segmap->doneData();

	if (rize) {
		for (i=0; i<nseg; i++) {
			float a1=frand(0,PI), a2=frand(0,2.0*PI);
			float mx[9] = {  cos(a2) , cos(a1)*sin(a2) , sin(a1)*sin(a2) ,
							-sin(a2) , cos(a1)*cos(a2) , sin(a1)*cos(a2) ,
							0 , -sin(a1) , cos(a1) };
			float tx,ty,tz;
			
			tx=(seg[i].x-nx/2)*mx[0]+(seg[i].y-ny/2)*mx[1]+(seg[i].z-nz/2)*mx[2]+nx/2;
			ty=(seg[i].x-nx/2)*mx[3]+(seg[i].y-ny/2)*mx[4]+(seg[i].z-nz/2)*mx[5]+ny/2;
			tz=(seg[i].x-nx/2)*mx[6]+(seg[i].y-ny/2)*mx[7]+(seg[i].z-nz/2)*mx[8]+nz/2;

			seg[i].x=tx;
			seg[i].y=ty;
			seg[i].z=tz;
		}
	}
	
	data->zero();
	dd=data->getDataRO();

	for (z=0,k=0; z<nz; z++) {
		for (y=0; y<ny; y++) {
			for (x=0; x<nx; x++,k++) {
				for (i=0; i<nseg; i++) {
					dd[k]+=exp(-hypot3(x-seg[i].x,y-seg[i].y,z-seg[i].z)/(2.0*seg[i].s));
				}
			}
		}
	}
	data->doneData();
	data->writeMRC(gaussmod);
}

LOG(Ref,NULL,LOG_END,NULL);
}
